import contextlib
import os
import boto3
import json
import re
import structlog

from botocore.exceptions import ClientError
from typing import Optional
from .common import REGION as COMMON_REGION
from .lang_utils import a_or_an
from .misc_utils import PRINT, full_class_name, override_environ


logger = structlog.getLogger(__name__)


# Environment variable whose value is the name of a secret
# stored in AWS Secrets Manager that contains the global
# application configuration. This secret JSON is acquired
# and returned to the caller. Note that this API can behave differently
# across AWS Accounts.
GLOBAL_APPLICATION_CONFIGURATION = 'IDENTITY'


# TODO: Rethink whether this should be a constant. Maybe it should be a function of the identity_kind?
#       Maybe the caller should pass a default and this function should have none. -kmp 18-Jul-2022
LEGACY_DEFAULT_IDENTITY_NAME = 'dev/beanstalk/cgap-dev'


def get_identity_name(identity_kind: str = GLOBAL_APPLICATION_CONFIGURATION):
    return os.environ.get(identity_kind, LEGACY_DEFAULT_IDENTITY_NAME)


def identity_is_defined(identity_kind: str = GLOBAL_APPLICATION_CONFIGURATION) -> bool:
    return bool(os.environ.get(identity_kind))


def get_identity_secrets(identity_kind: str = GLOBAL_APPLICATION_CONFIGURATION) -> dict:
    """ Grabs application identity from the secrets manager.
        Looks for environment variable IDENTITY, which should contain the name of
        a secret in secretsmanager that is a JSON blob of core configuration information.
        Default value is current value in the test account. This name should be the
        name of the environment.
    """
    secret_name = get_identity_name(identity_kind)
    region_name = COMMON_REGION  # us-east-1, must match ECR/ECS Region, which also imports its default from .common
    session = boto3.session.Session(region_name=region_name)
    client = session.client(
        service_name='secretsmanager',
        region_name=region_name
    )

    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:  # leaving some useful debug info to help narrow issues
        if e.response['Error']['Code'] in [
            'DecryptionFailureException'  # SM can't decrypt the protected secret text using the provided KMS key.
            'InternalServiceErrorException',  # An error occurred on the server side.
            'InvalidParameterException',  # You provided an invalid value for a parameter.
            'InvalidRequestException',  # Invalid parameter value for the current state of the resource.
            'ResourceNotFoundException',
        ]:
            PRINT('Encountered a known exception trying to acquire secrets.')
        raise e
    else:
        # Decrypts secret using the associated KMS CMK.
        # Depending on whether the secret is a string or binary, one of these fields will be populated.
        if 'SecretString' in get_secret_value_response:
            identity_secrets = json.loads(get_secret_value_response['SecretString'])
        else:
            raise Exception('Got unexpected response structure from boto3')
    if not identity_secrets:
        raise Exception('Identity could not be found! Check the secret name.')
    elif not isinstance(identity_secrets, dict):
        raise Exception("Identity was not in expected dictionary form.")
    return identity_secrets


assume_identity = get_identity_secrets  # assume_identity is deprecated. Please rewrite


def apply_overrides(*, secrets: dict, rename_keys: Optional[dict] = None,
                    override_values: Optional[dict] = None) -> dict:
    if rename_keys:
        secrets = secrets.copy()
        for key, new_name in rename_keys.items():
            if new_name in secrets:
                raise ValueError(f"Cannot rename {key} to {new_name}"
                                 f" because {a_or_an(new_name)} attribute already exists.")
            if key not in secrets:
                raise ValueError(f"Cannot rename {key} to {new_name} because the secrets has no {key} attribute.")
            secrets[new_name] = secrets.pop(key)
    if override_values:
        secrets = dict(secrets, **override_values)
    return secrets


@contextlib.contextmanager
def assumed_identity_if(only_if: bool, *,
                        identity_kind: str = GLOBAL_APPLICATION_CONFIGURATION,
                        only_if_missing: Optional[str] = None,
                        require_identity: bool = False,
                        rename_keys: Optional[dict] = None,
                        override_values: Optional[dict] = None):
    with assumed_identity(identity_kind=identity_kind, only_if=only_if, only_if_missing=only_if_missing,
                          require_identity=require_identity, rename_keys=rename_keys, override_values=override_values):
        yield


@contextlib.contextmanager
def assumed_identity(*,
                     identity_kind: str = GLOBAL_APPLICATION_CONFIGURATION,
                     only_if: bool = True,
                     only_if_missing: Optional[str] = None,
                     require_identity: bool = False,
                     rename_keys: Optional[dict] = None,
                     override_values: Optional[dict] = None):
    """
    Assumes a given identity in a context.

    The rename_keys happen before the override_values.

    >>> apply_overrides(secrets={'x': 1, 'y': 2}, rename_keys={'x': 'ex', 'y': 'why'}, override_values={'x': 3, 'z': 9})
    {'ex': 1, 'why': 2, 'z': 9, 'x': 3}

    NOTE: This assumes that global_env_bucket is in the GAC and not otherwise,
          so it tests

    :param identity_kind: The name of the identity to assume (default 'IDENTITY')
    :param only_if: Whether to try assuming identity at all (default True)
    :param only_if_missing: The name of an environment variable that would only be in the GAC.
       Load the GAC only if this variable is not set. Default is None, meaning load unconditionally.
    :param require_identity: Whether to raise an error if the identity_kind environment variable is not bound.
    :param rename_keys: If present, a dictionary mapping keys to override keys. Default None.
    :param override_values: If present, a dictionary mapping keys to override values. Default None.

    """
    if only_if and (not only_if_missing or not os.environ.get(only_if_missing)):
        identity_name = get_identity_name(identity_kind=identity_kind)
        if identity_name:
            secrets = get_identity_secrets(identity_kind=identity_kind)
            secrets = apply_overrides(secrets=secrets, rename_keys=rename_keys, override_values=override_values)
            if only_if_missing and only_if_missing not in secrets:
                raise RuntimeError(f"No {only_if_missing} was found where expected"
                                   f" in {identity_kind} secrets at {identity_name}.")
            with override_environ(**secrets):
                yield
                return  # Nothing to do after that

        elif require_identity:
            raise RuntimeError(f"An identity was not defined in environment variable {identity_kind}.")

    yield


def apply_identity(identity_kind: str = GLOBAL_APPLICATION_CONFIGURATION,
                   rename_keys: Optional[dict] = None, override_values: Optional[dict] = None):
    """
    Assumes an identity globally by assigning environment variables.

    Since this is assumed to be an initial startup action, we assume it's not already been done and don't
    offer the only_if_missing mechanism that the assumed_identity context manager offers.

    """
    if identity_is_defined(identity_kind):
        secrets = get_identity_secrets(identity_kind=identity_kind)
        secrets = apply_overrides(secrets=secrets, rename_keys=rename_keys, override_values=override_values)
        for var, val in secrets.items():
            os.environ[var] = val
    else:
        raise RuntimeError(f"An identity was not defined in environment variable {identity_kind}.")


# Some hints about how to manipulate secrets are here:
# https://docs.aws.amazon.com/code-samples/latest/catalog/python-secretsmanager-secretsmanager_basics.py.html
# However, the examples are poorly written and involve way too much weird state. -kmp 16-Feb-2022

class SecretsTable:
    """
    Assists with certain read-only operations on a given secret in the SecertsManager.

    This can be created with the name of what the AWS SecretsManager calls a secret,
    and what we call an identity. The idea is to allow read-only access to that data
    in tabular form.

    Once created, this table can be accessed via .as_dict() to turn it into a regular
    dictionary, but it can also be accessed via secrets_table.get(entryname) or
    secretstable[entryname] to access individual entries in the table. See tests for examples.

    NOTE: We could later extend this to allow creating or modifying the table as well,
    but that might involve extra permissions and be less broadly useful.
    """

    REGION = COMMON_REGION

    def __init__(self, name, secretsmanager_client=None, stage=None, region=None):
        """
        :param name: The SecretId of the secret to help with.
        :param secretsmanager_client: A Boto3 Secrets Manager client to use. If unsupplied, one is created.
        """
        if not str or not isinstance(name, str):
            raise ValueError(f"Bad secret name: {name}")
        self.secretsmanager_client = (secretsmanager_client
                                      or boto3.client('secretsmanager', region_name=region or self.REGION))
        self.name = name
        self.stage = stage

    def __str__(self):
        """
        Returns a string summarizing the table by name.

        :return: a string for use by str to describe this object as <SecretsTable 'name' @nnnn>
        """
        return f"<{full_class_name(self)} {self.name!r} @{'%x' % id(self)}>"

    def __repr__(self):
        """
        Returns a non-rereadable string summarizing the table by name.

        :return: a string for use by repr to describe this object as <SecretsTable 'name' @nnnn>,
                 the same as str does.
        """
        return self.__str__()

    def _get_secret_value(self):
        """
        Gets the 'value' of a secret in the SecretsManager.

        Even once a string is obtained, it may be the string containing the printed representation of a dictionary,
        so it may be more useful to use .get_secret_dict or to use .get or [...].

        :return: The value of the secret. When the secret is a string, the value is
                 contained in the `SecretString` field. When the secret is bytes,
                 it is contained in the `SecretBinary` field.
        """
        try:
            kwargs = {'SecretId': self.name}
            if self.stage is not None:
                kwargs['VersionStage'] = self.stage
            response = self.secretsmanager_client.get_secret_value(**kwargs)
        except ClientError as e:
            logger.error(f"Unable to retrieve value for secret {self.name}. {type(e)}: {e}")
            raise
        else:
            return response

    def _get_secret_string(self):
        """
        The secrets held by the AWS SecretsManager are actually strings or binary blobs.
        The strings are often JSON representing detailed entries in a table. This just
        gets the big string, in the case that it's string data (the only case we care about)
        so is mostly an internal thing. No user should be calling this.

        Note that this MIGHT return None if there is no string secret, usually because there was
        binary data instead. We expect that in our case there always will be string data in SecretString,
        but it would be in SecretBinary instead if there was binary data.
        """
        result = self._get_secret_value().get('SecretString')
        if result is None:
            raise ValueError(f"There is no SecretString for named {self.name!r}.")
        return result

    def as_dict(self):
        """
        Returns a representation of the secret as a Python dict.

        Note that calling this function gives the equivalent to what secrets_utils.assume_identity returns.
        See the tests for an example.

        :return: a dict representing the secrets table
        """
        secret_string = self._get_secret_string()
        if secret_string:
            try:
                return json.loads(secret_string)
            except Exception:
                raise ValueError(f"Unable to parse SecretString named {self.name} as a dictionary.")

    def get(self, item, default=None):
        """
        Returns the value of a given entry in the SecretsTable, or a default value (default None) otherwise.

        :item: a string naming the entry.
        :default: a default to use if the entry is missing (default None)

        :return: a string or None
        """
        return self.as_dict().get(item, default)

    def __getitem__(self, item):
        """
        Returns the value of a given entry in the SecretsTable, which is presumed to exist (or an error is raised).

        :item: a string naming the entry.

        :return: a string

        :raises KeyError: if the entry does not exist
        """
        result = self.get(item)
        if not result:
            raise KeyError(item)
        else:
            return result

    @classmethod
    def _all_secrets(cls, *, secretsmanager_client):
        """
        The AWS SecretsManager is itself a table. This gets a list of metadata dictionaries for each item in the
        SecretsManager, each of which can be the name of what we're here calling a SecretsTable. The name of
        each such metadata dictionary, the name of the associated secret, is in the dict entry called 'Name'.

        :return: a list of metadata dictionaries
        """
        return secretsmanager_client.list_secrets()['SecretList']

    @classmethod
    def _find_secrets(cls, *, pattern, secretsmanager_client):
        """
        This is like ._all_secrets() but returns a list of metadata dictionaries that correspond only to those
        whose names match the given regexp pattern.

        :return: a list of metadata dictionaries
        """
        results = []
        for secret in cls._all_secrets(secretsmanager_client=secretsmanager_client):
            if re.search(pattern, secret['Name']):
                results.append(secret)
        return results

    @classmethod
    def _find_secret(cls, *, pattern, secretsmanager_client):
        """
        This is like ._all_secrets() but returns exactly one metadata dictionary (not in a list)
        that has a name matching the given regexp pattern. It is an error if there are too many
        or none at all.

        :return: a metadata dictionary

        :raises RuntimeError: if zero matching dictionaries or more than one matching dictionaries are found
        """
        results = cls._find_secrets(secretsmanager_client=secretsmanager_client, pattern=pattern)
        n = len(results)
        if n == 0:
            raise RuntimeError(f"No secret was found matching {pattern!r}.")
        elif n == 1:
            return results[0]
        else:
            raise RuntimeError(f"Too many secrets match {pattern!r}: {[result['Name'] for result in results]}")

    @classmethod
    def find_secret(cls, pattern, secretsmanager_client=None):
        """
        This returns metadata about a secret with a name matching the given pattern.

        NOTE: This is like ._find_secret() but is not subprimitive. It is intended for calling by client interfaces,
        so it has nicer defaulting of arguments.

        :return: a metadata dictionary

        :raises RuntimeError: if zero matching dictionaries or more than one matching dictionaries are found
        """

        secretsmanager_client = secretsmanager_client or boto3.client('secretsmanager')
        return cls._find_secret(secretsmanager_client=secretsmanager_client, pattern=pattern)

    @classmethod
    def find_application_secrets_table(cls, secretsmanager_client=None,
                                       application_configuration_pattern='ApplicationConfiguration'):
        """
        This tries to find a secret representing an ApplicationConfiguration, since often there is only one.
        If there are multiple ApplicationConfigurations present, it may be necessary to specify a more specific
        search pattern via application_configuration_pattern=.

        :param secretsmanager_client: A boto3 client for secretsmanager. If one is not suppllied, it will be created.
           But it is sometimes good to reuse an existing one if you have one handy.
        :param application_configuration_pattern: A search pattern to use if there is intended to be be more than one
           application configuration to select among. The default should be fine if there is only one, although
           this somewhat depends on how 4dn-cloud-infra orchestration names things.

        :return: a SecretsTable representing the SecretString in the secret that is found by the search
        """
        secretsmanager_client = secretsmanager_client or boto3.client('secretsmanager')
        secret = cls.find_secret(pattern=application_configuration_pattern,
                                 secretsmanager_client=secretsmanager_client)
        secret_name = secret['Name']
        return SecretsTable(name=secret_name, secretsmanager_client=secretsmanager_client)



from typing_extensions import Literal


# ===== Useful constants =====

REGION = 'us-east-1'

APP_CGAP = 'cgap'
APP_FOURFRONT = 'fourfront'

LEGACY_GLOBAL_ENV_BUCKET = 'foursight-test-envs'
LEGACY_CGAP_GLOBAL_ENV_BUCKET = 'foursight-cgap-envs'

DEFAULT_ECOSYSTEM = 'main'

ORCHESTRATED_APPS = [APP_CGAP, APP_FOURFRONT]

CHALICE_STAGE_DEV = 'dev'
CHALICE_STAGE_PROD = 'prod'

CHALICE_STAGES = [CHALICE_STAGE_DEV, CHALICE_STAGE_PROD]

# ===== Type hinting names =====

EnvName = str

# Nicknames for enumerated sets of symbols. Note that these values must be syntactic literals,
# so they can't use the variables defined above.

ChaliceStage = Literal['dev', 'prod']
OrchestratedApp = Literal['cgap', 'fourfront']



import boto3
import logging
import os
import re

from .common import DEFAULT_ECOSYSTEM
from .lang_utils import conjoined_list
from .misc_utils import PRINT, find_associations, find_association, snake_case_to_camel_case, ignored
from .env_utils import is_stg_or_prd_env, ecr_repository_for_env


logging.basicConfig()
logger = logging.getLogger(__name__)


def camelize(snake_name):
    """
    Converts a token from hyphen-separated pseudo-snake-case to CamelCase.
    This is similar to snake_case_to_camel_case, but treats the common situation of hyphens
    that comes up in CloudFormation orchestration.
    """
    return snake_case_to_camel_case(snake_name, separator='-')


def dehyphenate(s):
    """Removes hyphens from a given string."""
    return s.replace('-', '')


def hyphenify(s):
    """Turns the underscore form of snake_case into the hyphenated form."""
    return s.replace('_', '-')


def tokenify(s):
    """Strip out everything but alphanumerics from a given token"""
    return ''.join([ch for ch in s if ch.isalpha() or ch.isdigit()])


def make_required_key_for_ecs_application_url(env_name):
    """
    This pattern needs to remain fixed for various things to connect up.
    Use make_key_for_ecs_application_url to generate an appropriate key.
    """
    return f"ECSApplicationURL{dehyphenate(env_name)}"


def get_ecs_real_url(env):
    """
    Returns the ECS target URL for a given environment created in on orchestration from Cloudformation
    if proper conventions were followed in by tagging the environment's ECS stack with an output
    generated by make_required_key_for_ecs_application_url.

    NOTE: Users of this facility are advised that they should not rely on printing a message and returning ''
    in the case of failure. In the future, an error may be raised, even without a major version bump.
    The only advertised behavior for now is that you should only call this when you expect success,
    and you should be prepared for a bad value OR raised errorin the case of failure.
    """
    manager = C4OrchestrationManager()
    found = manager.find_stack_output(make_required_key_for_ecs_application_url(env), value_only=True)
    if found:
        return found
    else:  # Couldn't this just raise an error? -kmp 17-Aug-2021
        PRINT('Did not locate the server from Cloudformation! Check ECS Stack metadata.')
        return ''


# This is the default ecosystem that the orchestration is expected to make, although it might be rational
# in a shared environment situation to have several ecosystems with names like 'test', 'prod', 'demo', etc.
# Things that in-ecosystem sharing could allow:
#
#  * Shared use of an ES engine, separated only by namespace, in a 'test' ecosystem,
#    even as another 'prod' ecosystem and its own ES (whether or not it had its own account).
#    Probably we would do such sharing only for Fourfront, since CGAP security requirements
#    will require better isolation.
#
# * Shared use of common repos between various platforms without repeated uploads.

def get_ecr_repo_url(env_name=None, *, ecosystem=DEFAULT_ECOSYSTEM, from_stack_output_key=None):
    """
    Gets the ECR repository URL of the repository from which ECS pulls application image runtimes from.

    NOTE: Users of this facility are advised that they should not rely on printing a message and returning ''
    in the case of failure. In the future, an error may be raised, even without a major version bump.
    The only advertised behavior for now is that you should only call this when you expect success,
    and you should be prepared for a bad value OR raised errorin the case of failure.
    """
    ignored(from_stack_output_key)  # This argument was once needed but is no longer.
    manager = C4OrchestrationManager()
    found = manager.find_stack_outputs(lambda x: x.endswith("RepoURL"), value_only=True)

    # Pick the most specific place the environment might belong...

    # If there is a specific match to the env, use that.
    env_name_suffix = f"/{env_name}"
    for url in found:
        if url.endswith(env_name_suffix):
            return url

    # If there was not repo called 'fourfront-blue' or 'fourfront-green',
    # but there is a 'fourfront-webprod' use that. This could be helpful
    # some day if there is an orchestrated cgap, but in the interim it can
    # still be used to find the fourfront-production ecr repo for fourfront.
    if is_stg_or_prd_env:
        prod_bucket_suffix = f"/{ecr_repository_for_env(env_name)}"
        for url in found:
            if url.endswith(prod_bucket_suffix):
                return url

    # Finally, if there is no match for any of those, look for an ecosystem-appropriate result.
    ecosystem_suffix = f"/{ecosystem}"
    for url in found:
        if url.endswith(ecosystem_suffix):
            return url

    # At this point, we have to give up.
    # Should we really not err? This just copies the return value convention we had previously. -17-Aug-2021
    PRINT('Did not locate the Repo URL from Cloudformation! Check ECR Stack metadata.')
    return ''


class AbstractOrchestrationManager:
    """This class consolidates some packaged functionality for osing around in in cloudformation."""

    COMMON_STACK_PREFIX = None  # Must be customized in a subclass. This is an abstract class.

    STACK_NAME_TOKEN_EXTRACTION_PATTERN = re.compile("^(.*)$")

    def __init__(self):
        self.cloudformation = boto3.resource('cloudformation')

    @classmethod
    def get_stack_output(cls, stack, output_key_name_or_predicate):
        """Utility function to get a specified output out of a given stack."""
        entry = find_association(stack.outputs or [], OutputKey=output_key_name_or_predicate)
        if not entry:
            return None
        return entry['OutputValue']

    def all_stacks(self):
        """Returns a list of all stacks whose names start with our COMMON_STACK_PREFIX (a class variable)."""
        return [
            stack
            for stack in self.cloudformation.stacks.all()
            if stack.name.startswith(self.COMMON_STACK_PREFIX)
        ]

    def _extract_stack_name_token(self, stack):
        """Internal function. Our name-related searches need to match a name that this function stracts."""
        m = self.STACK_NAME_TOKEN_EXTRACTION_PATTERN.match(stack.name)
        if m:
            return m.group(1)

    def find_stack(self, name_token):
        """
        Searches for a stack whose name matches the specified name_token.

        The name_token is not the full name, and is a bit more sophisticated than just a substring.
        This glosses over the fact that stacks have names like c4-network-trial-alpha-stack or
        c4-network-trial-stack, and allows us to see a particular part of the actual name as the
        name to search for. See STACK_NAME_TOKEN_EXTRACTION_PATTERN.
        """
        candidates = []
        for stack in self.all_stacks():
            if name_token == self._extract_stack_name_token(stack):
                candidates.append(stack)
        [candidate] = candidates or [None]
        return candidate

    def find_stack_outputs(self, key_or_pred, value_only=False):
        """
        If value_only is False, this finds all outputs of all stacks (see .all_stacks() for what "all" means),
        that match key_or_pred, returning a dictionary of {key1: val1, key2: val2, ...}.

        If value_only is True, then a list, rather than a dictionary, of the result values is returned.
        This would be odd in general, but serves cases where the result might be
        {"Mask1": "10.0.0.0", "Mask2": "192.168.0.0"} and where ["10.0.0.0", "192.168.0.0"] is preferred.

        NOTE WELL: We make the strong assumption about cross-stack output tag uniqueness, that all keys are distinct.
        If two stacks had the same output key, only one of them would be required.
        """
        results = {}
        for stack in self.all_stacks():
            for found in find_associations(stack.outputs or [], OutputKey=key_or_pred):
                results[found['OutputKey']] = found['OutputValue']
        if value_only:
            return list(results.values())
        else:
            return results

    def find_stack_output(self, key_or_pred, value_only=False):
        """
        This is like find_stack_outputs, but assumes that only zero or one values will be returned.
        If one value, it returns the first of what find_stack_outputs would return as a list.
        If no value, it returns None.
        """
        results = self.find_stack_outputs(key_or_pred, value_only=value_only)
        n = len(results)
        if n == 0:
            return None
        elif n == 1:
            return (results[0]  # in this case, result is a list, so take its first element
                    if value_only
                    else results)   # if not value-only, result is a dictionary, which is fine
        else:
            raise ValueError(f"Too many matches for {key_or_pred}: {results}")

    def find_stack_resource(self, stack_name_token, resource_logical_id, attr=None, default=None):
        """
        Looks up a resource or resource attribute in a given stack.

        This is intended to use our mechanisms for looking up a stack in spite of its complex name,
        and then returning a given resource or some attribute of it. For example:

          ConfigManager.find_stack_resource('foursight', 'CheckRunner', 'physical_resource_id')

        might find a stack named 'c4-foursight-trial-alpha-stack' and then would look up
        the physical_resource_id of its CheckRunner resource. Without the final argument, the
        resource itself is returned.

        The default is returned if the stack doesn't exist, the named resource doesn't exist,
        or (if a resource attribute is requested) no such attribute exists in the resource.
        """
        stack = self.find_stack(stack_name_token)
        if not stack:
            return default
        for summary in stack.resource_summaries.all():
            if summary.logical_id == resource_logical_id:
                if attr is None:
                    return summary
                else:
                    return getattr(summary, attr, default)
        return default

    CHECK_RUNNER_DEV_PATTERN = re.compile(".*foursight.*development.*CheckRunner.*")
    CHECK_RUNNER_PROD_PATTERN = re.compile(".*foursight.*production.*CheckRunner.*")

    ENCACHE_RUNNER_NAME = True

    @classmethod
    def discover_foursight_check_runner_name(cls, stage, encache=ENCACHE_RUNNER_NAME):

        check_runner = os.environ.get('CHECK_RUNNER')

        if check_runner:
            return check_runner

        # Prod has its own check runner, distinct from dev and test, though .get_stage() will have converted
        # 'test' to 'dev' by this point anyway. Still, this code is tolerant of not doing that...
        name_pattern = cls.CHECK_RUNNER_PROD_PATTERN if stage == 'prod' else cls.CHECK_RUNNER_DEV_PATTERN
        lambda_client = boto3.client('lambda')
        candidates = []
        chunk = lambda_client.list_functions()
        while True:
            entries = chunk['Functions']
            for entry in entries:
                name = entry['FunctionName']
                if name_pattern.match(name):
                    candidates.append(name)
            next_marker = chunk.get('NextMarker')
            if not next_marker:
                break
            chunk = lambda_client.list_functions(Marker=next_marker)
        if len(candidates) == 1:
            check_runner = candidates[0]
            logger.warning(f"CHECK_RUNNER inferred to be {check_runner}")
        else:
            logger.error(f"CHECK_RUNNER cannot be inferred"
                         f" from {conjoined_list(candidates, nothing='no matches')}.")

        if encache:
            # Discovery is slow. We could encache this value so it's faster next time. I defaulted this to off for now.
            logger.warning(f"Setting environment variable CHECK_RUNNER={check_runner}.")
            os.environ['CHECK_RUNNER'] = check_runner

        return check_runner


class C4OrchestrationManager(AbstractOrchestrationManager):
    """This will manage stacks like we use in the CGAP orchestration."""

    COMMON_STACK_PREFIX = 'c4-'

    STACK_NAME_TOKEN_EXTRACTION_PATTERN = re.compile(f"^{re.escape(COMMON_STACK_PREFIX)}([^-]*)(?:-.*|)$")


class AwsebOrchestrationManager(AbstractOrchestrationManager):
    # NOTE: This is not needed by CGAP orchestration, but illustrates how to generalize this facility for other uses.

    COMMON_STACK_PREFIX = 'awseb-e-'

    STACK_NAME_TOKEN_EXTRACTION_PATTERN = re.compile(f"^{re.escape(COMMON_STACK_PREFIX)}(.*)-stack$")



import boto3
import base64

from .common import REGION as COMMON_REGION
from .misc_utils import PRINT


class ECRUtils(object):
    """ Utility class for interacting with ECR.
        Initialized with an env name, from which a repo URL is resolved.

        NOTE 1: unlike s3Utils, the resolved resources (repository) will NOT be
        created if it does not already exist.
        NOTE 2: the (already created) ECR repository must have the env_name in the URI
    """

    REGION = COMMON_REGION  # this default must match what ecs_utils.ECSUtils and secrets_utils.assume_identity use

    # Will thinks this is no longer used. -kmp 14-Jul-2022
    #
    # ECR_LAYOUT = {
    #     'latest-wsgi': 'Latest version of WSGI Application',
    #     'latest-indexer': 'Latest version of Indexer Application',
    #     'latest-ingester': 'Latest version of the Ingester Application',
    #     'stable-wsgi': 'Stable version of WSGI Application',
    #     'stable-indexer': 'Stable version of Indexer Application',
    #     'stable-ingester': 'Stable version of the Ingester Application'
    # }

    # defaults were formerly env_name='cgap-mastertest', local_repository='cgap-wsgi'
    def __init__(self, *, env_name, local_repository, region=None):
        """ Creates an ECR client on startup """
        self.env = env_name
        self.local_repository = local_repository
        self.client = boto3.client('ecr', region_name=region or self.REGION)
        self.url = None  # set by calling the below method

    def resolve_repository_uri(self, url=None):
        if not self.url or url:
            PRINT('NOTE: Calling out to ECR')
            try:
                resp = self.client.describe_repositories()
                for repo in resp.get('repositories', []):
                    if repo['repositoryUri'].endswith(self.env):
                        url = repo['repositoryUri']
            except Exception as e:
                PRINT('Could not retrieve repository information from ECR: %s' % e)
        self.url = url  # hang onto this
        return url

    def get_uri(self):
        """ Returns URI if it's been set, raise exception otherwise """
        if self.url is not None:
            return self.url
        raise Exception('Tried to get URI when it has not been resolved yet!')

    def authorize_user(self):
        """ Calls to boto3 to get authorization credentials for ECR.
            Gives a result like this:
                {
                    'authorizationData': [{'authorizationToken': 'very long token....',
                                            'expiresAt': <datetime>,
                                            'proxyEndpoint': <ecr URL> ... }]
                }
        """
        try:
            [auth_data] = self.client.get_authorization_token()['authorizationData']
            return auth_data
        except Exception as e:
            PRINT('Could not acquire ECR authorization credentials: %s' % e)
            raise

    @staticmethod
    def extract_ecr_password_from_authorization(*, authorization):
        """ Extracts the password from the authorization token returned from the
            above API call.
        """
        return base64.b64decode(authorization['authorizationToken']).replace(b'AWS:', b'').decode('utf-8')


# These two variables are deprecated. Please use references to ECRUtils instead. They will go away in the future.
# CGAP_ECR_LAYOUT = ECRUtils.ECR_LAYOUT
CGAP_ECR_REGION = ECRUtils.REGION



import boto3
# import botocore.client
import contextlib
# import json
# import logging
import os
# import urllib.request

# from typing import Optional
from .common import LEGACY_GLOBAL_ENV_BUCKET
from .exceptions import (
    # CannotInferEnvFromManyGlobalEnvs,
    # MissingGlobalEnv,
    SynonymousEnvironmentVariablesMismatched, LegacyDispatchDisabled,
)
from .misc_utils import (
    override_environ,
    # ignored,
    remove_suffix,
)


class LegacyController:
    LEGACY_DISPATCH_ENABLED = False


class EnvBase:

    @classmethod
    def global_env_bucket_name(cls):
        """
        This class method returns the name of the current 'global env bucket', the bucket where meanings of
        environment names are looked up in orchestrated environments.
        """
        global_bucket_env_var = 'GLOBAL_BUCKET_ENV'  # Deprecated. Supported for now since some tools started using it.
        global_env_bucket_var = 'GLOBAL_ENV_BUCKET'  # Preferred name. Please transition code to use this.
        global_bucket_env = os.environ.get(global_bucket_env_var)
        global_env_bucket = os.environ.get(global_env_bucket_var)
        if global_env_bucket and global_bucket_env and global_env_bucket != global_bucket_env:
            raise SynonymousEnvironmentVariablesMismatched(var1=global_bucket_env_var, val1=global_bucket_env,
                                                           var2=global_env_bucket_var, val2=global_env_bucket)
        global_bucket = global_env_bucket or global_bucket_env
        return global_bucket

    @classmethod
    @contextlib.contextmanager
    def global_env_bucket_named(cls, name):
        """
        This class method, a 'context manager' useful with the Python 'with' operation, binds the name of
        the current 'global env bucket', the bucket where meanings of environment names are looked up
        in orchestrated environments.
        """

        with override_environ(GLOBAL_BUCKET_ENV=name, GLOBAL_ENV_BUCKET=name):
            yield

    @classmethod
    def _legacy_global_env_bucket_for_testing(cls):
        if not LegacyController.LEGACY_DISPATCH_ENABLED:
            raise LegacyDispatchDisabled(operation="_legacy_global_env_bucket_for_testing", mode='setup-envbase')
        # Strictly speaking, this isn't accessing the legacy state, but it's definitely not accessin gproduction state,
        # so it does not want to be used in production. -kmp 18-Jul-2022
        return LEGACY_GLOBAL_ENV_BUCKET

    @classmethod
    def _get_configs(cls, env_bucket, kind):
        env_bucket_name = (
            # prefer a given bucket
            env_bucket
            # or GLOBAL_ENV_BUCKET
            or cls.global_env_bucket_name()
            # but failing that, for legacy system, just use legacy name
            or cls._legacy_global_env_bucket_for_testing())
        s3_resource = boto3.resource('s3')
        env_bucket_model = s3_resource.Bucket(env_bucket_name)
        key_names = [key_obj.key for key_obj in env_bucket_model.objects.all()]
        configs = cls.filter_config_names(key_names, kind=kind)
        return sorted(configs)

    @classmethod
    def filter_config_names(cls, configs, kind):
        result = []
        for config in configs:
            if kind == 'env' and '.' not in config:
                result.append(config)
            elif kind == 'ecosystem' and config.endswith('.ecosystem'):
                result.append(remove_suffix('.ecosystem', config))
        return result

    @classmethod
    def get_all_environments(cls, env_bucket=None):
        return cls._get_configs(env_bucket=env_bucket, kind='env')

    @classmethod
    def get_all_ecosystems(cls, env_bucket=None):
        return cls._get_configs(env_bucket=env_bucket, kind='ecosystem')


class s3Base:

    # Some extra variables used in setup here so that other modules can be consistent with chosen values.

    SYS_BUCKET_SUFFIX = "system"
    OUTFILE_BUCKET_SUFFIX = "wfoutput"
    RAW_BUCKET_SUFFIX = "files"
    BLOB_BUCKET_SUFFIX = "blobs"
    METADATA_BUCKET_SUFFIX = "metadata-bundles"
    TIBANNA_OUTPUT_BUCKET_SUFFIX = 'tibanna-output'
    TIBANNA_CWLS_BUCKET_SUFFIX = 'tibanna-cwls'

    # NOTE: These were deprecated and retained for compatibility in dcicutils 2.
    #       For dcicutils 3.0, please rewrite uses as HealthPageKey.xyz names.
    # SYS_BUCKET_HEALTH_PAGE_KEY = HealthPageKey.SYSTEM_BUCKET                     # = 'system_bucket'
    # OUTFILE_BUCKET_HEALTH_PAGE_KEY = HealthPageKey.PROCESSED_FILE_BUCKET         # = 'processed_file_bucket'
    # RAW_BUCKET_HEALTH_PAGE_KEY = HealthPageKey.FILE_UPLOAD_BUCKET                # = 'file_upload_bucket'
    # BLOB_BUCKET_HEALTH_PAGE_KEY = HealthPageKey.BLOB_BUCKET                      # = 'blob_bucket'
    # METADATA_BUCKET_HEALTH_PAGE_KEY = HealthPageKey.METADATA_BUNDLES_BUCKET      # = 'metadata_bundles_bucket'
    # TIBANNA_CWLS_BUCKET_HEALTH_PAGE_KEY = HealthPageKey.TIBANNA_CWLS_BUCKET      # = 'tibanna_cwls_bucket'
    # TIBANNA_OUTPUT_BUCKET_HEALTH_PAGE_KEY = HealthPageKey.TIBANNA_OUTPUT_BUCKET  # = 'tibanna_output_bucket'
    # This is also deprecated, even though not a bucket name. Use HealthPageKey.ELASTICSEARCH.
    # ELASTICSEARCH_HEALTH_PAGE_KEY = HealthPageKey.ELASTICSEARCH                  # = 'elasticsearch'

    EB_PREFIX = "elasticbeanstalk"
    EB_AND_ENV_PREFIX = EB_PREFIX + "-%s-"  # = "elasticbeanstalk-%s-"

    SYS_BUCKET_TEMPLATE = EB_AND_ENV_PREFIX + SYS_BUCKET_SUFFIX            # = "elasticbeanstalk-%s-system"
    OUTFILE_BUCKET_TEMPLATE = EB_AND_ENV_PREFIX + OUTFILE_BUCKET_SUFFIX    # = "elasticbeanstalk-%s-wfoutput"
    RAW_BUCKET_TEMPLATE = EB_AND_ENV_PREFIX + RAW_BUCKET_SUFFIX            # = "elasticbeanstalk-%s-files"
    BLOB_BUCKET_TEMPLATE = EB_AND_ENV_PREFIX + BLOB_BUCKET_SUFFIX          # = "elasticbeanstalk-%s-blobs"
    METADATA_BUCKET_TEMPLATE = EB_AND_ENV_PREFIX + METADATA_BUCKET_SUFFIX  # = "elasticbeanstalk-%s-metadata-bundles"
    TIBANNA_OUTPUT_BUCKET_TEMPLATE = TIBANNA_OUTPUT_BUCKET_SUFFIX          # = "tibanna-output" (no prefix)
    TIBANNA_CWLS_BUCKET_TEMPLATE = TIBANNA_CWLS_BUCKET_SUFFIX              # = "tibanna-cwls" (no prefix)
